#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import path, makedirs, remove
from sys import stdin, stdout
from shutil import rmtree, make_archive
from base64 import b64encode, b64decode
from uuid import uuid4
from zipfile import ZipFile

from unlocker.util.passkey import Passkey
from unlocker.util.log import Log


class Migrate(object):
    """Secrets migration in and out.

    Arguments:
        migrate_format  (str): The temporary file format of archive.
        migrate_tmpdir  (str): Temporary directory to unpack and pack secrets.
        migrate_tmpfile (str): Name of temporary file to unpack and pack.

    Raises:
        Exception: If manager raises any exceptions.
    """

    migrate_format = "zip"
    migrate_tmpdir = "/tmp/.x-unlocker"
    migrate_tmpfile = "secrets.list"

    def __init__(self, manager):
        self.manager = manager

    def import_secrets(self):
        """Import secrets wrapper.

        Read and decode from stdin and import authority and passkeys.
        """

        filename = self.get_tmp_filename()
        filepath = self.get_tmp_path(filename)
        with open(filepath, "wb") as fd:
            fd.write(b64decode(stdin.read()))
        with ZipFile(filepath, "r") as zf:
            self.read_tmp_secrets(zf)

    def read_tmp_secrets(self, zf_secrets):
        """Read temporary secrets.

        Args:
            zf_secrets (ZipFile): ZipFile instance with secrets and passkeys.

        Raises:
            SystemExit: If manager crashes or corrupted data are found.
        """

        for each in zf_secrets.read(self.migrate_tmpfile).split("\n"):
            line = each.split("\t")
            if len(line) == 0:
                continue
            _, _, host, ipv4, port, user, scheme, name, ptype, passkey = line
            if ptype not in Passkey.SUPPORTED_TYPES:
                Log.fatal("Unsupported passkey storage {x}", x=ptype)
            authority_args = user, host, port, scheme
            auth = self.manager.build_authority_from_args(*authority_args)
            if ptype == "privatekey":
                passkey = zf_secrets.read(passkey)
            data = {
                "name": name,
                "host": host,
                "auth": auth,
                "passkey": Passkey.SUPPORTED_TYPES.get(ptype) + passkey
            }
            self.manager.get_db().add(**data)
        Log.warn("Unsupported import for jump server, yet")

    def export_secrets(self):
        """Export secrets wrapper.

        Loop through all secrets and export authority and passkeys.

        Raises:
            SystemExit: If there's nothing to export.
        """

        rows = []
        for name, auth, host, jump in self.manager.get_db().query_all():
            ipv4, port = auth.get_host_ip4(), str(auth.get_port())
            user, scheme = auth.get_user(), auth.get_scheme()
            _, _, secret = self.manager.get_db().lookup(name)
            passtype, passkey = Passkey.copy(secret, True)
            if passtype == "privatekey":
                passkey = self.dump_pk_file(passkey, user, host, scheme)
            jump_auth = "n/a"
            if jump is not None:
                jump_auth = jump.signature()
            rows.append((
                auth.signature(),  # auth signature
                jump_auth,         # jump signature if any
                host,              # hostname
                ipv4,              # IPv4 address
                port,              # port number
                user,              # username
                scheme,            # connection scheme
                name,              # authority name
                passtype,          # passkey type
                passkey            # actual passkey or path to passkey
            ))
        if len(rows) == 0:
            Log.fatal("Nothing to export...")
        self.flush_secrets(rows)

    def flush_secrets(self, content):
        """Flush secrets to stdout.

        It creates additional files for private keys whenever needed. Files
        are deleted after they are flushed to stdout.

        Args:
            content (str): Content with secrets to write to file.
        """

        # write all secrets to file...
        with open(self.get_tmp_path(self.migrate_tmpfile), "wb") as fd:
            fd.write("\n".join(["\t".join(r) for r in content]))

        # create temporary archive and flush it's content to stdout
        exp_pkg = "/tmp/{}".format(self.get_tmp_filename())
        make_archive(exp_pkg, self.migrate_format, self.migrate_tmpdir)
        with open("{}.{}".format(exp_pkg, self.migrate_format), "rb") as fd:
            stdout.write(b64encode(fd.read()))

        # cleanup
        remove("{}.{}".format(exp_pkg, self.migrate_format))

    def get_tmp_filename(self):
        """Temporary filename getter.

        Generates a filename for temporary storage.

        Returns:
            str: Temporary filename.
        """

        return str(uuid4().hex.upper())

    def get_tmp_path(self, filename):
        """Temporary path getter.

        Returns path to migration temporary directory.

        Args:
            filename (str): Name of file from temp directory.

        Returns:
            str: Path to file from temporary directory.
        """

        return "{}/{}".format(self.migrate_tmpdir, filename)

    def create_pk_file(self, user, hostname, scheme):
        """Create a private key filename.

        Args:
            user   (str): Username from Authority.
            host   (str): Hostname from Authority.
            scheme (str): Scheme from Authority.

        Returns:
            str: Filename for private key.
        """

        return "{}_{}_{}.pk".format(user, hostname, scheme)

    def dump_pk_file(self, passkey, user, host, scheme):
        """Write private key to file and return file name.

        Args:
            passkey (str): Passkey as private key to write to file.
            user    (str): Username from Authority.
            host    (str): Hostname from Authority.
            scheme  (str): Scheme from Authority.

        Returns:
            str: Name of file with private key.

        """

        filename = self.create_pk_file(user, host, scheme)
        with open(self.get_tmp_path(filename), "wb") as fd:
            fd.write(passkey)
        return filename

    @classmethod
    def discover(cls, manager):
        """
        Args:
            manager (Manager): Manager instance.

        Raises:
            SystemExit: If invalid options are provided.
        """

        import_secrets = manager.args.get("import_secrets")
        export_secrets = manager.args.get("export_secrets")

        # fail fast if no options is provided...
        if not import_secrets and not export_secrets:
            Log.fatal("Unexpected migrate request...")

        # create tmp dir if not eixsts...
        if not path.exists(cls.migrate_tmpdir):
            makedirs(cls.migrate_tmpdir)

        # run migration
        mig = cls(manager)
        if import_secrets:
            if stdin.isatty():
                Log.fatal("Migration failed: stdin is empty")
            mig.import_secrets()
        elif manager.args.get("export_secrets"):
            if stdout.isatty():
                Log.fatal("Migration failed: stdout is empty")
            mig.export_secrets()

        # cleanup
        rmtree(cls.migrate_tmpdir)



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from sys import argv, stdin
from re import compile, UNICODE

from unlocker.authority import Authority

from unlocker.util.service import Service


AUTHORITY_REGEXPR = "(?:(?P<scheme>.+)://)" \
                    "(?:(?P<user>.+?)@)" \
                    "(?P<host>[\da-zA-Z\-\.]+)" \
                    "(?:\:(?P<port>\d+))?"


class StreamData(object):
    """Standard stream data wrapper.

    Arguments:
        buf_in (str): Input buffer used to store piped stdin.
    """

    buf_in = None
    OPTION = "stdout_dump"

    @classmethod
    def read(cls):
        """Check if authority is piped.

        Returns:
            bool: True if data is piped, otherwise False.
        """

        if len(argv) > 1:
            return False
        if not stdin.isatty():
            cls.buf_in = unicode(stdin.read()).strip()
            return len(cls.buf_in) > 0
        return False

    @classmethod
    def parse(cls):
        """Parse input buffer.

        Returns:
            dict: Dictionary with arguments.
        """

        args = {"name": "", "signature": ""}
        exp = compile(AUTHORITY_REGEXPR, UNICODE)
        matches = exp.match(cls.buf_in)
        if matches is None:
            args.update({"name": cls.buf_in})
        else:
            params = matches.groupdict()
            if params.get("port") is None:
                params["port"] = Service.find_port(params.get("scheme"))
            auth = Authority.new(**params)
            args.update({"signature": auth.signature()})
        return args



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from stat import S_IEXEC
from os import path, makedirs, stat, chmod
from os.path import expanduser
try:
    from os import geteuid

    def is_root(): return geteuid() == 0
except ImportError:
    def is_root(): return False

from unlocker.util.log import Log

from unlocker import __version__, __project__


SYSTEM_SCRIPTS_DIR = "/usr/local/bin"
SCRIPTS_DIR = "bin"
SCRIPTS_SHEBANG = ur"""#!/bin/sh
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

VERSION={version}
HOMEPAGE={homepage}
""".format(version=__version__, homepage="http://github.com/lexndru/unlocker")


def deploy_lock_script():
    """Install LOCK script on current machine.

    Raises:
        Exception: If script is missing or cannot be accessed.

    Return:
        bool: True if successfully deployed, otherwise False.
    """

    return deploy_script("lock.sh")


def deploy_unlock_script():
    """Install UNLOCK script on current machine.

    Raises:
        Exception: If script is missing or cannot be accessed.

    Return:
        bool: True if successfully deployed, otherwise False.
    """

    return deploy_script("unlock.sh")


def deploy_script(script_name):
    """Install script on current machine.

    Args:
        script_name (str): The name of script to be deployed.

    Raises:
        Exception: If script is missing or cannot be accessed.

    Return:
        bool: True if successfully deployed, otherwise False.
    """

    content = read_helper_script("data/shell/{}".format(script_name))
    if not content:
        Log.fatal("Helper script is missing")
    script = "{}\n{}".format(SCRIPTS_SHEBANG, content)
    try:
        script_name, _ = script_name.split(".", 1)
    except Exception:
        Log.debug("Cannot find file extension for {s}", s=script_name)
    return make_helper_script(script_name, script)


def read_helper_script(filepath):
    """Read helper script

    Args:
        filepath (str): Path to helper script to read.

    Raises:
        Exception: If helper script cannot be accessed.

    Return:
        unicode: Unicode content of helper script.
    """

    try:
        file_script = path.abspath(path.join(__project__, filepath))
        if not path.exists(file_script):
            Log.fatal("Cannot find helper script {file}", file=filepath)
        with open(file_script, "rb") as fd:
            return fd.read()
    except Exception as e:
        Log.fatal("Cannot read helper script: {e}", e=str(e))


def make_helper_script(file_script, script_content):
    """Create executable helper script.

    Args:
        file_script        (str): Script to make.
        script_content (unicode): Script content.

    Raises:
        Exception: If application cannot create directory or file.

    Return:
        bool: True if script was successfully created, otherwise False.
    """

    if is_root():
        scripts_dir = SYSTEM_SCRIPTS_DIR
    else:
        scripts_dir = path.join(expanduser("~"), SCRIPTS_DIR)
    Log.debug("Scripts directory located at {path}", path=scripts_dir)
    try:
        if not path.exists(scripts_dir):
            Log.debug("Scripts directory does not exist. Creating...")
            makedirs(scripts_dir)
            Log.debug("Scripts directory successfully created!")
        else:
            Log.debug("Found scripts directory...")
    except Exception as e:
        Log.fatal("Failed to create scripts directory: {err}", err=str(e))
    try:
        filepath = "{}/{}".format(scripts_dir, file_script)
        if path.exists(filepath):
            Log.fatal("A file with this name already exists: {x}", x=filepath)
        with open(filepath, "wb") as fd:
            fd.write(script_content)
        mod = stat(filepath)
        chmod(filepath, mod.st_mode | S_IEXEC)
        return True
    except Exception as e:
        Log.fatal("Cannot create helper script because: {e}", e=str(e))
    return False



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from uuid import uuid4

from unlocker.authority import Authority
from unlocker.keychain import Keychain
from unlocker.database import Database
from unlocker.migrate import Migrate
from unlocker.display import Display

from unlocker.util.service import Service
from unlocker.util.passkey import Passkey
from unlocker.util.log import Log


class Manager(object):
    """Authority secrets manager.

    The Manager is a dispatcher for various call options, such as saving
    passkeys and removing them or lookup.

    There are four possible keys to be accessed by the manager:
      1) storage keys: used to store passkeys (passwords or private keys);
      2) authority keys: used to save authority instances;
      3) host keys: used to keep track of the hostname of an authority;
      4) jump keys: used to keep track of relationships between authorities.

    It's mandatory for the manager to always orchestrate storage, authority and
    host keys. Without there three, the keychain is considered to be corrupted.

    A saved passkey is prefixed with the passkey's type in a "storage key"
    where a storage key is a known named authority.
      let "storage_key" be "storage_prefix" + "named authority"
      e.g. storage_key = passkey_type + actual_passkey

    Authority keys are similar to storage_key, except the content stored is
    compacted by the Authority layer.
      let "auth_key" be "auth_prefix" + "named authority"
      e.g. auth_key = component_authority_formatted_string

    Hostnames are saved in hostnames keys as plain strings.
      let "hostname_key" be "hostname_prefix" + "named authority"
      e.g. hostname_key = localhost

    The optional jump authority keys are saved in jump keys pointing to another
    authority from the keychain. These jump authorities are used for tunneling
    or bouncing.
      let "jump_key" be "jump_prefix" + "named authority"
      e.g. jump_key = another_component_authority_formatted_string

    Arguments:
        __secrets (Keychain): Authority and passwords database.

    Args:
        option     (str): Manager operation to handle.
        args      (dict): Arguments provided to manager.
        auth (Authority): Current authority working with.
    """

    __secrets, __database = None, None

    supported_options = {
        # option        access level
        "migrate":      "read_write",
        "append":       "read_write",
        "update":       "read_write",
        "remove":       "read_write",
        "forget":       "read_write",
        "lookup":       "read_only",
        "recall":       "read_only",
        "list":         "read_only",
        "dump":         "debug_read",
        "purge":        "debug_write",
        "stdout_dump":  "secret_read",
    }

    MAX_ITER_LIST = 2**16
    MIN_NAME_LEN, MAX_NAME_LEN = 10, 42  # meaning of life

    def __init__(self, option, args):
        self.option = option
        self.args = args
        self.auth = None

    @classmethod
    def initialize(cls, secrets):
        """Register keychain to manager's database.

        Args:
            secrets (object): Dict-like object storage.
        """

        cls.__secrets = Keychain(secrets)
        cls.__database = Database(cls.__secrets)
        Log.debug("Manager initialized...")

    def get_secrets(self):
        """Keychain getter.

        Returns:
            Keychain: Passkeys storage.
        """

        if isinstance(self.__secrets, Keychain):
            return self.__secrets
        Log.fatal("Missing keychain: manager not initialized?")

    def get_db(self):
        """Database getter.

        Returns:
            Database: Database storage.
        """

        if isinstance(self.__database, Database):
            return self.__database
        Log.fatal("Missing database: manager not initialized?")

    def build_authority_from_args(self, user, host, port=None, scheme=None):
        """Generate authority from a arguments.

        Args:
            user   (str): The username to attach to authority.
            host   (str): The hostname to resolve and attach to authority.
            port   (int): The port number to attach to authority.
            scheme (str): The connection scheme.

        Raises:
            Exception: If arguments are invalid.

        Returns:
            Authority: The generated authority instance.
        """

        if port is None and scheme is not None:
            port = Service.find_port(scheme)
        if port is not None and scheme is None:
            scheme = Service.find_scheme(port)
        return Authority.new(host, port, user, scheme)

    def build_authority_from_signature(self, signature):
        """Generate authority from a signature.

        Args:
            signature (str): The signature to generate the authority.

        Raises:
            Exception: If no authority matches the signature provided.

        Returns:
            Authority: The authority with the signature provided.
        """

        for each, _ in self.get_db().query_auth():
            if each.signature() == signature:
                return each
        Log.fatal("Cannot find authority for signature {s}", s=signature)

    def build_random_name(self):
        """Generate a random name with 16 characters.

        Returns:
            str: Random 16 characters.
        """

        name = uuid4().hex[:16]
        Log.debug("Generating new random name: {n}", n=name)
        return name

    def call(self):
        """Call dispatcher.

        Handlers option if is identified and supported.

        Raises:
            Exception: If unsupported option is provided.
        """

        if self.option not in self.supported_options:
            Log.fatal("Unsupported option {o}", o=self.option)
        access = self.supported_options[self.option]
        debug_message = "Calling option {opt} with access level {acc}"
        Log.debug(debug_message, opt=self.option, acc=access)
        method = "call_{}_{}_option".format(access, self.option)
        if not hasattr(self, method):
            error = "Unsupported method {option} or access rights {access}"
            Log.fatal(error, option=self.option, access=access)
        return getattr(self, method)(**self.args)

    def call_read_only_recall_option(self, signature, **kwargs):
        """Lookup wrapper without explicit authority.

        Finds authority by signature and calls "lookup" for passkey.

        Args:
            key (str): Possible authority signature to match or name.

        Raises:
            Exception: If named authority does not exists.

        Outputs:
            stdout: Human-readable passkey with authority details and hostname.
        """

        Log.debug("Searching secret passkey for key {s}", s=signature)
        if len(signature) < self.MIN_NAME_LEN:
            Log.debug("Trying key as authority signature...")
            for name, auth, _, _ in self.get_db().query_all():
                if auth.signature() == signature:
                    Log.debug("Got authority with signature {s}", s=signature)
                    signature = name
                    break
        Log.debug("Running a lookup for named authority: {n}", n=signature)
        self.call_read_only_lookup_option(signature)

    def call_read_write_forget_option(self, signature, **kwargs):
        """Remove wrapper without explicit authority.

        Finds authority by signature and calls "remove" for passkey.

        Args:
            key (str): Possible authority signature to match or name.

        Raises:
            Exception: If named authority does not exists.

        Outputs:
            stdout: Human-readable passkey with authority details and hostname.
        """

        Log.debug("Testing if key {s} exists", s=signature)
        if len(signature) < self.MIN_NAME_LEN:
            Log.debug("Trying key as authority signature...")
            for name, auth, _, _ in self.get_db().query_all():
                if auth.signature() == signature:
                    Log.debug("Got authority with signature {s}", s=signature)
                    signature = name
                    break
        Log.debug("Running cleanup after named authority: {n}", n=signature)
        self.call_read_write_remove_option(signature)

    def call_read_write_update_option(self, name, auth, jump_server=None,
                                      **kwargs):
        """Update secrets for an existing named authority.

        Args:
            name (str): The name of the authority to lookup.
            auth (str): Authentification method.

        Raises:
            Exception: If named authority does not exists.

        Outputs:
            stdout: Confirm message with human-readable authority details.
        """

        Log.debug("Incoming update request for named authority {n}", n=name)
        if not self.get_db().exists(name):
            Log.debug("Named authority is not found...")
            error = "Cannot update entry: \"{name}\" not found in " \
                    "keychain (missing key)"
            Log.fatal(error, name=name)
        if jump_server is not None:
            Log.debug("Update requests to change jump server...")
            jump_auth = self.build_authority_from_signature(jump_server)
            self.get_db().update_jump_auth(name, jump_auth)
            Log.debug("New jump set to authority: {a}", a=str(jump_auth))
        passkey = Passkey.resolve(auth)
        self.get_db().update_passkey(name, passkey)
        Log.debug("New passkey set ... ")
        Display.show_update(self.get_db().fetch_auth(name))

    def call_read_write_append_option(self, name, host, port, user, auth,
                                      scheme, jump_server, **kwargs):
        """Append secrets to a named authority.

        Args:
            name        (str): The name of the authority to lookup.
            host        (str): Hostname to attach to authority.
            port        (int): Port number to attach to authority.
            user        (str): Username to attach to authority.
            auth        (str): Authentification method.
            scheme      (str): Scheme of the connection.
            jump_server (str): Name or signature of another authority.

        Raises:
            Exception: If named authority already exists.

        Outputs:
            stdout: Confirm message with human-readable authority details.
        """

        Log.debug("Incoming append request...")
        if name is None:
            name = self.build_random_name()
            Log.warn("Named authority is known as: {n}", n=name)
        elif len(name) < self.MIN_NAME_LEN:
            error = "Possible lookup collision: entry name is too short " \
                    "(name must be at least {size} characters)"
            Log.fatal(error, size=self.MIN_NAME_LEN)
        elif len(name) > self.MAX_NAME_LEN:
            error = "Possible bad naming: entry name is too long " \
                    "(name must be at most {size} characters)"
            Log.fatal(error, size=self.MAX_NAME_LEN)
        elif self.get_db().exists(name):
            error = "Another entry with the same name exists: {name} " \
                    "(name must be unique)"
            Log.fatal(error, name=name)
        data = {
            "name": name,
            "host": host,
            "auth": self.build_authority_from_args(user, host, port, scheme),
        }
        if jump_server is not None:
            data.update({
                "jump_auth": self.build_authority_from_signature(jump_server)
            })
        Log.debug("Preparing to add {args}", args=data)
        self.get_db().add(passkey=Passkey.resolve(auth), **data)
        Log.debug("New named authority is saved...")
        Display.show_append(data.get("auth"))

    def call_read_write_remove_option(self, name, **kwargs):
        """Lookup and remove secrets for a named authority.

        Args:
            name (str): The name of the authority to lookup.

        Raises:
            Exception: If named authority does not exists.

        Outputs:
            stdout: Human-readable passkey with authority details and hostname.
        """

        Log.debug("Incoming remove request for named authority {n}", n=name)
        if not self.get_db().exists(name):
            Log.debug("Nothing to remove...")
            error = "Cannot remove entry: \"{name}\" not found in " \
                    "keychain (missing key)"
            Log.fatal(error, name=name)
        Log.debug("Fetching data for named authority...")
        auth, host, secret = self.get_db().lookup(name)
        dependents = []
        for jump, dep_name in self.get_db().query_jump():
            if jump.signature() == auth.signature():
                debug_message = "Found another authority {a} " \
                                "depending on this... "
                Log.debug(debug_message, a=str(jump))
                dependents.append(dep_name)
        if len(dependents) > 0:
            error = "Not removing entry because {n} other servers bounce of " \
                    "\"{name}\": remove all before trying again (safe mode)"
            Log.fatal(error, n=len(dependents), name=name)
        passtype, passkey = Passkey.copy(secret, True)
        Log.debug("Removing named authority...")
        self.get_db().remove(name)  # remove all keys for named auth
        Log.debug("Permanently removed named authority {n}", n=name)
        Display.show_remove(auth, host, passtype, passkey)

    def call_read_only_lookup_option(self, name, **kwargs):
        """Lookup secrets for a named authority.

        Args:
            name (str): The name of the authority to lookup.

        Raises:
            Exception: If named authority does not exists.

        Outputs:
            stdout: Human-readable passkey with authority details and hostname.
        """

        Log.debug("Incoming lookup request for named authority {n}", n=name)
        if not self.get_db().exists(name):
            Log.debug("Nothing to lookup...")
            error = "Cannot lookup entry: \"{name}\" not found in " \
                    "keychain (missing key)"
            Log.fatal(error, name=name)
        Log.debug("Fetching data for named authority...")
        auth, host, secret = self.get_db().lookup(name)
        passtype, passkey = Passkey.copy(secret, True)
        Log.debug("Printing as safe as possible {n}'s secrets...", n=name)
        Display.show_lookup(auth, host, passtype, passkey)

    def call_read_only_list_option(self, *args, **kwargs):
        """List handler.

        Outputs:
            stdout: Pager with table-like view of all hostnames.
        """

        Log.debug("Incoming list request...")
        known_hosts = [e for e in self.get_db().query_all()]
        counter = -1
        indexes = {}
        sorted_hosts = []
        Log.debug("Found {n} hosts to list...", n=len(known_hosts))
        while len(known_hosts) > 0:
            if counter >= self.MAX_ITER_LIST:
                Log.fatal("Max. iteration over list display reached...")
            name, auth, host, jump = known_hosts[0]
            counter += 1
            if auth.signature() not in indexes:
                indexes.update({auth.signature(): counter})
            if jump is None:
                sorted_hosts.append(known_hosts.pop(0))
                continue
            jump_index = indexes.get(jump.signature(), -1)
            if jump_index > -1:
                if jump_index == 0:
                    jump_index += 1
                sorted_hosts.insert(jump_index, known_hosts.pop(0))
                continue
            known_hosts.insert(len(known_hosts), known_hosts.pop(0))
        Log.debug("Sorted all known hosts and now preparing to print out...")
        Display.show_list_view(sorted_hosts, **self.args)

    def call_read_write_migrate_option(self, *args, **kwargs):
        """Migration wrapper.

        Determine if action is to import or to export and launch process.

        Raises:
            Exception: If something goes terrible wrong.

        Inputs:
            stdin: Compressed import-ready secrets.

        Outputs:
            stdout: Compressed exported stored secrets.
        """

        Log.debug("Incoming migrate request...")
        Migrate.discover(self)
        Log.debug("Migrated data...")

    def call_debug_read_dump_option(self, keys):
        """Dump all keys from keychain in debug mode.
        """

        Log.debug("Incoming debug dump request...")
        for key in keys:
            for k in self.get_secrets().lookup(key):
                key_type = self.get_db().which(k)
                Log.debug(" [{t}] {k}", k=self.get_db().shift(k), t=key_type)
        Log.debug("Closing...")

    def call_debug_write_purge_option(self, keys):
        """Permanently remove keys from keychain in debug mode.

        Can corrupt keychain! (Very unsafe)
        """

        Log.debug("Incoming debug purge request...")
        for key in keys:
            Log.debug("Attempt to remove {k} ...", k=key)
            removed_key = self.get_secrets().remove(key)
            if removed_key is not None:
                Log.debug("Removed {k} ...", k=key)
        Log.debug("Closing...")

    def call_secret_read_stdout_dump_option(self, name, signature, **kwargs):
        """Vulnerable passkey dump to stdout.

        Args:
            name      (str): The name of the authority to lookup.
            signature (str): The signature for an authority to find its name.

        Outputs:
            stdout: Base64 encoded passkey.

        Raises:
            Exception: If unsupported passkey is found.
        """

        def print_passkey(auth_name):
            _, _, secret = self.get_db().lookup(auth_name)
            return "{}\n{}".format(*Passkey.copy(secret))

        passkey = ""  # dummy passkey
        Log.debug("Incoming secret dump request...")
        if len(name) > 0:
            Log.debug("Got named authority {n}...", n=name)
            passkey = print_passkey(name)
        elif len(signature) > 0:
            for each, name in self.get_db().query_auth():
                if each.signature() != signature:
                    continue
                Log.debug("Records matched authority, got passkey...")
                passkey = print_passkey(name)
                break  # exit on first match
        Log.debug("Printing passkey..." if passkey else "Nothing to print...")
        Display.show_dump(passkey)



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import environ
from sys import stdout

try:
    if environ.get("NOPAGER", "") == "true":
        raise ImportError("NOPAGER is set")
    from click import echo_via_pager as print_page
except ImportError:
    def print_page(content): stdout.write(content)

from unlocker.util.log import Log


APPEND_TEMPLATE = u"""
  Secrets successfully added!

  {user}@{host}:{port}

""".encode("utf-8")

UPDATE_TEMPLATE = u"""
  Secrets successfully updated!

  {user}@{host}:{port}

""".encode("utf-8")

REMOVE_TEMPLATE = u"""
  Permanently removed secret {auth_type} for {host}!

  \033[91mThis is the last chance to save this secret.\033[0m

  {user}@{ipv4}:{port}    \x1b[0;37;47m{passkey}\x1b[0m

""".encode("utf-8")

LOOKUP_TEMPLATE = u"""
  Secret {auth_type} for {host}

  {user}@{ipv4}:{port}    \x1b[0;37;47m{passkey}\x1b[0m

""".encode("utf-8")

VERTICAL_LIST_TEMPLATE = u"""
{nr:>4}) {name} ({sig}{jump_server})
      Hostname: {host}
      IPv4: {ip}
      Port: {port} ({proto})
      User: {user}
""".encode("utf-8")


class Display(object):
    """Pretty display manager.

    Used to display messages on read/write actions triggered by Manager.
    """

    DEFAULT_MESSAGE = "Nothing to show..."
    LINE_SEPARATOR = "\n"

    @classmethod
    def show(cls, content=None):
        """Display content in pager or flush to stdout.

        Args:
            content (mixt): Content to output.

        Raises:
            Exception: If unsupported content is provided.

        Output:
            stdout: Content via pager if supported.
        """

        if content is None:
            content = cls.DEFAULT_MESSAGE
        if isinstance(content, (list, tuple, set)):
            content = cls.LINE_SEPARATOR.join(content)
        if not isinstance(content, (str, unicode)):
            Log.fatal("Cannot display non-string content")
        print_page(content)

    @classmethod
    def show_lookup(cls, auth, host, pass_type, passkey):
        """Display passkey for lookup message.

        Args:
            auth (Authority): Matched authority.
            host       (str): Matched hostname.
            pass_type  (str): Authentification method.
            passkey    (str): Password or private key.
        """

        if pass_type == "privatekey":
            passkey = "\n" + passkey
        content = LOOKUP_TEMPLATE.format(
            ipv4=auth.get_host_ip4(), port=auth.get_port(), host=host,
            user=auth.get_user(), auth_type=pass_type, passkey=passkey)
        cls.show(content)

    @classmethod
    def show_remove(cls, auth, host, pass_type, passkey):
        """Display passkey for last time in remove message.

        Args:
            auth (Authority): Matched authority.
            host       (str): Matched hostname.
            pass_type  (str): Authentification method.
            passkey    (str): Password or private key.
        """

        content = REMOVE_TEMPLATE.format(
            ipv4=auth.get_host_ip4(), port=auth.get_port(), host=host,
            user=auth.get_user(), auth_type=pass_type, passkey=passkey)
        cls.show(content)

    @classmethod
    def show_append(cls, auth):
        """Display confirmation on append message.

        Args:
            auth (Authority): Matched authority.
        """

        content = APPEND_TEMPLATE.format(
            host=auth.get_host_ip4(), port=auth.get_port(),
            user=auth.get_user())
        cls.show(content)

    @classmethod
    def show_update(cls, auth):
        """Display confirmation on update message.

        Args:
            auth (Authority): Matched authority.
        """

        content = UPDATE_TEMPLATE.format(
            host=auth.get_host_ip4(), port=auth.get_port(),
            user=auth.get_user())
        cls.show(content)

    @classmethod
    def show_list_view_vertical(cls, rows, **kwargs):
        """Display records from keychain in less than 80 chars per line.

        Args:
            rows (list): Records from keychain.
        """

        content = []
        for name, auth, host, jump in rows:
            jump_server = ""
            if jump is not None:
                jump_server = " => {}".format(jump.signature())
            record = VERTICAL_LIST_TEMPLATE.format(
                       sig=auth.signature(), host=host, ip=auth.get_host_ip4(),
                       port=auth.get_port(), proto=auth.get_scheme(),
                       user=auth.get_user(), nr=len(content) + 1,
                       name=name, jump_server=jump_server)
            content.append(record)
        cls.show(content)

    @classmethod
    def show_list_view(cls, rows, column, exclude, vertical, **kwargs):
        """Display records from keychain in a table-like view.

        Args:
            rows     (list): Records from keychain.
            columns  (list): Columns to display.
            exclude  (bool): Whether to exclude columns or not.
            vertical (bool): Whether to display in compatibility mode or not.
        """

        if vertical:
            ignore_message = "Ignoring {} options in compatibility mode"
            ignore_items = []
            if len(column) > 0:
                ignore_items.append("columns")
            if exclude:
                ignore_items.append("exclude")
            if len(ignore_items) > 0:
                Log.warn(ignore_message.format("/".join(ignore_items)))
            return cls.show_list_view_vertical(rows)
        headers = {
            "auth_sig": "hash",
            "jump_sig": "bounce",
            "ip4": "ipv4",
            "port": "port",
            "user": "user",
            "proto": "protocol",
            "host": "hostname",
            "name": "friendly name",
        }
        records = []
        max_host_len = len(headers["host"])
        max_user_len = len(headers["user"])
        max_name_len = len(headers["name"])
        for name, auth, host, jump in rows:
            if len(name) > max_name_len:
                max_name_len = len(name)
            if len(host) > max_host_len:
                max_host_len = len(host)
            if len(auth.get_user()) > max_user_len:
                max_user_len = len(auth.get_user())
            records.append({
                "auth_sig": auth.signature(),
                "jump_sig": jump.signature() if jump is not None else "~",
                "ip4": auth.get_host_ip4(),
                "port": auth.get_port(),
                "user": auth.get_user(),
                "proto": auth.get_scheme(),
                "host": host,
                "name": name,
                "jump": jump is not None
            })
        line = ["{proto:^8}", "{ip4:^15}", "{port:^5}"]
        line.append("{host:^%s}" % max_host_len)
        line.append("{user:^%s}" % max_user_len)
        line.append("{name:^%s}" % max_name_len)
        tpl = u" {auth_sig:^10} | {jump_sig:^10} | " + u" | ".join(line)
        line_tpl = tpl.encode("utf-8")
        content = [line_tpl.format(**headers)]
        separator = {}
        counters = {
            "auth_sig": 10,
            "jump_sig": 10,
            "ip4": 15,
            "port": 5,
            "user": max_user_len,
            "proto": 8,
            "name": max_name_len,
            "host": max_host_len
        }
        for k, v in counters.iteritems():
            separator[k] = unicode(v * "=")
        content.append(line_tpl.format(**separator))
        for record in records:
            content.append(line_tpl.format(**record))
        content.append(cls.LINE_SEPARATOR)
        if len(records) == 0:
            content = ["Nothing to display... try \"append\" or \"update\""]
        cls.show(content)

    @classmethod
    def show_dump(cls, passkey_dump):
        """Vulnerable passkey dump to stdout.

        Args:
            passkey_dump (str): Base64 passkey dump with type as prefix.

        Outputs:
            stdout: Base64 encoded passkey.
        """

        stdout.write(passkey_dump.strip())



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

from os import environ
from sys import argv
from argparse import ArgumentParser

from unlocker.util.log import Log
from unlocker.util.secret import Secret
from unlocker.util.helper import deploy_unlock_script, deploy_lock_script

from unlocker import __version__


OKAY_MESSAGE = "OK"
HELP_MESSAGE = """
              _            _
  _   _ _ __ | | ___   ___| | _____ _ __
 | | | | '_ \| |/ _ \ / __| |/ / _ \ '__|
 | |_| | | | | | (_) | (__|   <  __/ |
  \__,_|_| |_|_|\___/ \___|_|\_\___|_|

Unlocker v{} - CLI credentials manager

Usage:
  init          Create local keychain
  list          List known hosts from keychain
  recall        Retrieve secrets by name or signature (slower than lookup)
  forget        Forget secrets by name or signature (slower than remove)
  append        Add new set of credentials to keychain
  update        Update password or private key to keychain
  remove        Remove credentials from keychain
  lookup        Find password for provided host, port and user
  install       Install helper scripts
  migrate       Migrate secrets to current unlocker version
""".format(__version__)

SCRIPTS_CREATED = """OK
The following commands are now available:
  unlock  - Establish connection to known servers
  lock    - Encrypt secrets storage"""


class ShellArgumnets(object):

    name = ("-n", "--name"), {
        "help": "A name to recognize it later",
        "dest": "name",
        "action": "store"
    }

    host = ("-h", "--host"), {
        "help": "The hostname or IP address",
        "dest": "host",
        "action": "store"
    }

    port = ("-p", "--port"), {
        "help": "The port number to use for the connection",
        "dest": "port",
        "action": "store",
        "type": int
    }

    user = ("-u", "--user"), {
        "help": "The username to use for the connection",
        "dest": "user",
        "action": "store"
    }

    auth = ("-a", "--auth"), {
        "help": "The authentication method used",
        "dest": "auth",
        "action": "store",
        "choices": ["password", "privatekey"]
    }

    service = ("-s", "--scheme"), {
        "help": "The scheme service name to display (e.g. ssh, http, mysql)",
        "dest": "scheme",
        "action": "store",
    }

    jump_server = ("-j", "--jump-server"), {
        "help": "Optional jump server (or tunnel server) for connections",
        "dest": "jump_server",
        "action": "store",
    }


def get_append_shell(self, header="Add new set of credentials"):
    """Shell getter for "append" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "append" option.
    """

    arguments = self.build_args([
        ShellArgumnets.host, ShellArgumnets.port,
        ShellArgumnets.user, ShellArgumnets.auth
    ], True)
    optional_arguments = self.build_args([
        ShellArgumnets.name, ShellArgumnets.service, ShellArgumnets.jump_server
    ])
    arguments.update(optional_arguments)
    return self.get_parser(arguments, header).parse_args(argv[2:])


def get_update_shell(self, header="Update set of credentials"):
    """Shell getter for "update" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "update" option.
    """

    arguments = self.build_args([
        ShellArgumnets.name, ShellArgumnets.auth
    ], True)
    optional_arguments = self.build_args([
        ShellArgumnets.jump_server
    ])
    arguments.update(optional_arguments)
    return self.get_parser(arguments, header).parse_args(argv[2:])


def get_remove_shell(self, header="Remove credentials from keys database"):
    """Shell getter for "remove" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "remove" option.
    """

    arguments = self.build_args([ShellArgumnets.name], True)
    return self.get_parser(arguments, header).parse_args(argv[2:])


def get_lookup_shell(self, header="Find key for provided user@host:port"):
    """Shell getter for "lookup" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "lookup" option.
    """

    arguments = self.build_args([ShellArgumnets.name], True)
    return self.get_parser(arguments, header).parse_args(argv[2:])


def get_recall_shell(self, header="Retrieve passkey from keychain"):
    """Shell getter for "recall" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "recall" option.
    """

    psr = ArgumentParser(description=header)
    psr.add_argument("signature", help="Authority signature to lookup")
    return psr.parse_args(argv[2:])


def get_forget_shell(self, header="Permanently forget passkey"):
    """Shell getter for "forget" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "forget" option.
    """

    psr = ArgumentParser(description=header)
    psr.add_argument("signature", help="Authority signature to remove")
    return psr.parse_args(argv[2:])


def get_list_shell(self, header="List known hosts from keychain"):
    """Shell getter for "list" option.

    Args:
        header (str): Description header to display on help message.

    Returns:
        Namespace: Parsed arguments namespace for "list" option.
    """

    psr = ArgumentParser(description=header)
    psr.add_argument(
        "-e", "--exclude", action="store_true", dest="exclude",
        help="Exclude columns given as positional arguments")
    psr.add_argument(
        "-v", "--vertical", action="store_true", dest="vertical",
        help="Display list of hosts vertically (80 columns compatibility)")
    psr.add_argument("column", help="Columns to display", nargs="*")
    return psr.parse_args(argv[2:])


def get_init_shell(self):
    """Shell getter for "init" option.
    """

    try:
        Secret.get_secret_file()
        print(OKAY_MESSAGE)
    except Exception as e:
        Log.fatal("Aborting due to an error: {e}", e=str(e))
    raise SystemExit


def get_install_shell(self):
    """Shell getter for "install" option.
    """

    try:
        deploy_unlock_script() and deploy_lock_script()
        print(SCRIPTS_CREATED)
    except Exception as e:
        Log.fatal("Aborting due to an error: {e}", e=str(e))
    raise SystemExit


def get_migrate_shell(self):
    """Shell getter for "migrate" option.
    """

    if len(argv[2:]) > 0:
        psr = ArgumentParser(description="Migrate stored secrets")
        grp = psr.add_mutually_exclusive_group()
        grp.add_argument("--import",
                         action="store_true",
                         dest="import_secrets",
                         help="Import secrets from STDIN")
        grp.add_argument("--export",
                         action="store_true",
                         dest="export_secrets",
                         help="Export secrets to STDOUT")
        return psr.parse_args(argv[2:])
    try:
        Secret.migrate_secrets()
        print(OKAY_MESSAGE)
    except Exception as e:
        Log.fatal("Aborting due to an error: {e}", e=str(e))
    raise SystemExit


def get_dump_shell(self):
    """Shell getter for "dump" option.

    Option available only in debug mode.
    """

    psr = ArgumentParser(description="Dump list of keys (debug mode)")
    psr.add_argument("keys", help="Keys to debug (e.g. host)")
    return psr.parse_args(argv[2:])


def get_purge_shell(self):
    """Shell getter for "purge" option.

    Option available only in debug mode.
    """

    psr = ArgumentParser(description="Drop keys from storage (debug mode)")
    psr.add_argument("keys", help="Keys to drop from keychain", nargs="+")
    return psr.parse_args(argv[2:])


methods = {
    "get_init_shell": get_init_shell,
    "get_list_shell": get_list_shell,
    "get_recall_shell": get_recall_shell,
    "get_forget_shell": get_forget_shell,
    "get_append_shell": get_append_shell,
    "get_update_shell": get_update_shell,
    "get_remove_shell": get_remove_shell,
    "get_lookup_shell": get_lookup_shell,
    "get_install_shell": get_install_shell,
    "get_migrate_shell": get_migrate_shell,
}

if "DEBUG" in environ:
    methods.update({
        "get_dump_shell": get_dump_shell,
        "get_purge_shell": get_purge_shell
    })
    HELP_MESSAGE += """
Debug:
  dump          Dump all entries from keychain
  purge         Force delete key from keychain (can corrupt entire keychain!)
"""

OptionParser = type("OptionParser", (object,), methods)


class ShellParser(OptionParser):
    """Command line arguments wrapper.

    Arguments:
        option (str): Given option to parse (append, remove, lookup).

    Raises:
        SystemExit: If unsupported option is provided.
    """

    def __init__(self):
        self.option = argv[1] if len(argv) > 1 else self.get_help()

    def get_help(self):
        """Print a help message and exit.

        Raises:
            SystemExit: After message is printed.
        """

        print(HELP_MESSAGE)
        raise SystemExit

    def get_args(self):
        """Call supported shell based on given option.

        Raises:
            SystemExit: If an unsupported option is given or not implemented.

        Returns:
            tuple: Option name and mamespace from the called shell.
        """

        try:
            shell = getattr(self, "get_{}_shell".format(self.option))
        except AttributeError:
            return self.get_help()
        if not callable(shell):
            return self.get_help()
        return self.option, shell()

    def get_parser(self, arguments, description="n/a"):
        """Parsed builder and getter based on given option.

        Args:
            arguments  (dict): Dictionary of supported optional arguments.
            description (str): Description to display on help message.

        Returns:
            Parser: Argument parser build for given option with arguments.
        """

        psr = ArgumentParser(description=description, add_help=False)
        for keys, vals in arguments.iteritems():
            short, long_ = keys
            if short is None:
                psr.add_argument(long_, **vals)
            else:
                psr.add_argument(short, long_, **vals)
        psr.add_argument("--help", action="help",
                         help="show this help message and exit")
        return psr

    def build_args(self, args=(), required=False):
        """Arguments builder helper.

        Args:
            args     (iter): Iterable arguments to build.
            required (bool): Whether fields are reqired or not.

        Returns:
            dict: Dictionary of arguments.
        """

        return {k: v.update({"required": required}) or v for k, v in args}



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from zlib import compress, decompress
from base64 import b64encode, b64decode

from unlocker.util.log import Log


class Keychain(object):
    """Credentials storage wrapper.

    Arguments:
        keychain (object): Storage dict-like object.

    Args:
        holder   (object): Storage instance or object.
    """

    def __init__(self, holder):
        self.keychain = holder
        Log.debug("Keychain initialized...")

    def add(self, key, value):
        """Append key to keychain.

        Args:
            key   (str): Key to append.
            value (str): Value to save for given key.
        """

        if self.has(key):
            Log.fatal("Cannot add duplicates in keychain")
        self.update(key, value)

    def has(self, key):
        """Check if keychain has given key.

        Args:
            key (str): Key to lookup

        Returns:
            bool: True if keychain has key, otherwise False.
        """

        for k in self.lookup(key, partial=False):
            return True
        return False

    def get_value(self, key):
        """Returns real value for given key.

        Args:
            key (str): Key to lookup and retrieve value.

        Raises:
            Exception: If value is None or keychain does not have key.

        Returns:
            str: Uncompressed and decoded stored value for key.
        """

        value = self.get(key)
        if value is None:
            Log.fatal("Keychain does not have requested key")
        return decompress(b64decode(value))

    def get(self, key):
        """Returns "as is" value for given key.

        Args:
            key (str): Key to lookup and retrieve value.

        Raises:
            Exception: If value is not string or keychain does not have key.

        Returns:
            str: Raw "as is" base64 stored value for key.
        """

        if not self.has(key):
            return None
        return self.keychain[key]

    def update(self, key, value):
        """Update key in keychain.

        Args:
            key   (str): Key to update.
            value (str): Value to save for given key.
        """

        self.keychain[key] = b64encode(compress(value))

    def remove(self, key):
        """Remove key from keychain.

        Args:
            key (str): Key to remove.

        Returns:
            str: Raw "as is" removed value for given key.
        """

        value = self.get(key)
        if value is None:
            Log.warn("Keychain can not remove an unset key")
        else:
            del self.keychain[key]
        return value

    def lookup(self, key, partial=True):
        """Lookup key in keychain.

        Args:
            key (str): Key to lookup.

        Yields:
            str: Yields exact key(s) if found.
        """

        if hasattr(self.keychain, "iterkeys"):
            iterator = self.keychain.iterkeys
        else:
            iterator = self.keychain.keys
        for k in iterator():
            if key == k or (partial and k.startswith(key)):
                yield k

    def __repr__(self):
        return "[{} key(s) stored]".format(len(self.keychain))



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from unlocker.keychain import Keychain
from unlocker.authority import Authority

from unlocker.util.secret import Secret
from unlocker.util.log import Log


class Database(object):
    """Database common interface.

    Supports high-level and quick access to keychain.
    """

    # used by version key
    VERSION = Secret.VERSION

    # used as second character after a key type prefix
    SEPARATOR = "!"

    # key type prefix (storage, auth, host, jump)
    PASS, AUTH, HOST, JUMP = "$", "A", "h", "j"

    # minimum length of a prefix with separator
    PREFIX_FIXED_LEN = 2

    def __init__(self, storage):
        if not isinstance(storage, Keychain):
            Log.fatal("Unexpected database storage {t}", t=type(storage))
        self.storage = storage
        Log.debug("Database initialized...")
        Log.debug("Storage status: {k}", k=str(storage))

    def exists(self, name):
        """Tests whether a named authority exists in keychain.

        Args:
            name (str): Full name of the authority to test.

        Returns:
            bool: True if named authority exists, otherwise False.
        """

        return self.storage.has(self.get_pass_key(name))

    def add_passkey(self, name, passkey):
        """Create new passkey for a named authority.

        Args:
            name    (str): Full name of the authority to add.
            passkey (str): Processed passkey to save to keychain.

        Raises:
            Exception: If named authority already exists in keychain.
        """

        if self.exists(name):
            Log.fatal("Cannot add passkey on a duplicate entry")
        self.update_passkey(name, passkey)

    def add_auth(self, name, auth):
        """Create new authority for self named authority.

        Args:
            name       (str): Full name of the authority to add.
            auth (Authority): Authority instance to save to keychain.

        Raises:
            Exception: If named authority already exists in keychain.
        """

        if self.storage.has(self.get_auth_key(name)):
            Log.fatal("Cannot add authority on a duplicate entry")
        if not isinstance(auth, Authority):
            Log.fatal("Expected auth to be authority, got {t}", t=type(auth))
        self.storage.add(self.get_auth_key(name), auth.read())

    def add_host(self, name, host):
        """Create new hostname for named authority.

        Args:
            name (str): Full name of the authority to add.
            host (str): Hostname to save to keychain.

        Raises:
            Exception: If named authority already exists in keychain.
        """

        if self.storage.has(self.get_host_key(name)):
            Log.fatal("Cannot add hostname on a duplicate entry")
        self.storage.add(self.get_host_key(name), host)

    def add_jump(self, name, auth):
        """Create new jump authority for named authority.

        Args:
            name       (str): Full name of the authority to add.
            auth (Authority): Jump authority instance to save to keychain.

        Raises:
            Exception: If named authority already exists in keychain.
        """

        if self.storage.has(self.get_jump_key(name)):
            Log.fatal("Cannot add jump server on a duplicate entry")
        if not isinstance(auth, Authority):
            Log.fatal("Expected jump to be authority, got {t}", t=type(auth))
        self.storage.add(self.get_jump_key(name), auth.read())

    def update_passkey(self, name, passkey):
        """Update passkey for existing named authority.

        Args:
            name    (str): Full name of the authority to add.
            passkey (str): New processed passkey to replace the old passkey.

        Raises:
            Exception: If passkey is zero-length.
        """

        if len(passkey) == 0:
            Log.fatal("Passkey cannot be empty")
        self.storage.update(self.get_pass_key(name), passkey)

    def update_jump_auth(self, name, auth):
        """Update passkey for existing named authority.

        Args:
            name       (str): Full name of the authority to add.
            auth (Authority): New authority as jump server.

        Raises:
            Exception: If provided argument is not Authority.
        """

        if not isinstance(auth, Authority):
            Log.fatal("Expected authority instance, got {t}", t=type(auth))
        self.storage.update(self.get_jump_key(name), auth.read())

    def add(self, name, passkey, auth, host=None, jump_auth=None):
        """Create entry for named authority.

        Args:
            name            (str): Full name of the authority to add.
            passkey         (str): Processed passkey to save to keychain.
            auth      (Authority): Authority instance to save to keychain.
            host            (str): Hostname to save to keychain.
            jump_auth (Authority): Jump authority instance to save to keychain.

        Raises:
            Exception: If any of the methods raise an exception.
        """

        self.add_passkey(name, passkey)
        self.add_auth(name, auth)
        if host is not None:
            self.add_host(name, host)
        if jump_auth is not None:
            self.add_jump(name, jump_auth)

    def remove_passkey(self, name):
        """Remove storage key containing passkey from keychain.

        Args:
            name (str): Full name of the authority to remove.

        Outputs:
            stdout: Prints a warning if the storage key is not found.

        Returns:
            mixt: The passkey just removed (if found) or None.
        """

        return self.storage.remove(self.get_pass_key(name))

    def remove_auth(self, name):
        """Remove authority key containing authority from keychain.

        Args:
            name (str): Full name of the authority to remove.

        Outputs:
            stdout: Prints a warning if the authority key is not found.

        Returns:
            mixt: The authority just removed (if found) or None.
        """

        return self.storage.remove(self.get_auth_key(name))

    def remove_host(self, name):
        """Remove hostname key containing hostname from keychain.

        Args:
            name (str): Full name of the authority to remove.

        Outputs:
            stdout: Prints a warning if the hostname key is not found.

        Returns:
            mixt: The hostname just removed (if found) or None.
        """

        return self.storage.remove(self.get_host_key(name))

    def remove_jump(self, name):
        """Remove jump authority key containing authority from keychain.

        Args:
            name (str): Full name of the authority to remove.

        Outputs:
            stdout: Prints a warning if the jump authority key is not found.

        Returns:
            mixt: The jump authority just removed (if found) or None.
        """

        return self.storage.remove(self.get_jump_key(name))

    def remove(self, name):
        """Remove all keys from keychain for a named authority.

        Args:
            name (str): Full name of the authority to remove.

        Outputs:
            stdout: Prints a warning if a key is not found.

        Raises:
            Exception: If any of the methods used raise an exception.
        """

        if self.storage.has(self.get_jump_key(name)):
            self.remove_jump(name)
        self.remove_host(name)
        self.remove_auth(name)
        self.remove_passkey(name)

    def fetch(self, name, query, key_name="name"):
        """Retieve entry from keychain for a named authority.

        Args:
            name       (str): Full name of the authority to fetch.
            query (callable): Query callable to filter for `name`.
            key_name   (str): Placeholder key for message to output.

        Raises:
            Exception: If entry is not found or query is not callable.

        Returns:
            mixt: Entry to retrieve.
        """

        if not callable(query):
            Log.fatal("Cannot fetch with an invalid query")
        for each, name_ in query():
            if name_ == name:
                return each
        Log.fatal("Cannot fetch unexisting {k}: {n}", n=name, k=key_name)

    def fetch_auth(self, name):
        """Retieve authority from keychain for a named authority.

        Args:
            name (str): Full name of the authority to fetch.

        Raises:
            Exception: If entry is not found or query is not callable.

        Returns:
            Authority: Authority to retrieve.
        """

        return self.fetch(name, self.query_auth, "authority")

    def fetch_host(self, name):
        """Retieve hostname from keychain for a named authority.

        Args:
            name (str): Full name of the authority to fetch.

        Raises:
            Exception: If entry is not found or query is not callable.

        Returns:
            str: Hostname to retrieve.
        """

        return self.fetch(name, self.query_host, "hostname")

    def fetch_jump(self, name):
        """Retieve jump authority from keychain for a named authority.

        Args:
            name (str): Full name of the authority to fetch.

        Raises:
            Exception: If entry is not found or query is not callable.

        Returns:
            Authority: Jump authority to retrieve.
        """

        return self.fetch(name, self.query_jump, "jump server")

    def query(self, key_type_prefix):
        """Query secrets from keychain storage.

        Args:
            key_type_prefix (str): Prefix to partial match in lookup.

        Yields:
            mixt: Entry for each matched found.
        """

        for each in self.storage.lookup(key_type_prefix):
            if each == self.VERSION:
                continue
            if self.storage.get_value(each).strip() == "":
                Log.fatal("Storage contains empty value for key {k}", k=each)
            yield each

    def query_auth(self):
        """Query authorities from keychain storage.

        Yields:
            tuple: Authority instance and the named key.
        """

        for each in self.query(self.get_auth_prefix()):
            name = self.shift(each)
            yield Authority.recover(self.storage.get_value(each)), name

    def query_host(self):
        """Query hostnames from keychain storage.

        Yields:
            tuple: Hostname and the named key.
        """

        for each in self.query(self.get_host_prefix()):
            name = self.shift(each)
            yield self.storage.get_value(each), name

    def query_jump(self):
        """Query jump authorities from keychain storage.

        Yields:
            tuple: Jump authority instance and the named key.
        """

        for each in self.query(self.get_jump_prefix()):
            name = self.shift(each)
            yield Authority.recover(self.storage.get_value(each)), name

    def query_all(self):
        """Query everything in relation to authority from keychain storage.

        Yields:
            tuple: The named key, authority instance, hostname and jump auth.
        """

        for each in self.query(self.get_pass_prefix()):
            if len(each) <= self.PREFIX_FIXED_LEN:
                continue
            name = self.shift(each)
            auth = self.storage.get_value(self.get_auth_key(name))
            authority = Authority.recover(auth)
            host = None
            if self.storage.has(self.get_host_key(name)):
                host = self.storage.get_value(self.get_host_key(name))
            jump_auth = None
            if self.storage.has(self.get_jump_key(name)):
                jump = self.storage.get_value(self.get_jump_key(name))
                jump_auth = Authority.recover(jump)
            yield name, authority, host, jump_auth

    def lookup(self, lookup_name):
        """Lookup a named authority and return self, hostname and secret.

        Args:
            lookup_name (str): The name of the authority to find.

        Raises:
            Exception: If authority is not found.

        Returns:
            tuple: Authority instance, hostname and secret passkey.
        """

        for name, auth, host, jump in self.query_all():
            if name == lookup_name:
                secret = self.storage.get_value(self.get_pass_key(name))
                return auth, host, secret
        Log.fatal("Nothing found for name {n}", n=lookup_name)

    def shift(self, string):
        """Shift to the right a sting to remove any prefix.

        Args:
            string (str): String to be shifted.

        Raises:
            Exeption: If provided argument is not a string.

        Returns:
            str: New shifted string without prefix.
        """

        if isinstance(string, (str, unicode)):
            return string[self.PREFIX_FIXED_LEN:]
        Log.fatal("Unsupported shift operation on {t}", t=type(string))

    def get_prefix(self, prefix):
        """Prefix getter. Appends a separator and returns prefix.

        Args:
            prefix (str): The prefix to return.

        Raises:
            Exception: If size of prefix exceeds.

        Returns:
            str: New string of prefix with separator.
        """

        key = prefix + self.SEPARATOR
        if len(key) != self.PREFIX_FIXED_LEN:
            Log.fatal("Invalid prefixed key length, got {v}", v=len(key))
        return key

    def get_pass_key(self, key):
        """Passkey generator and getter.

        Args:
            key (str): Key to format and return with prefix.

        Returns:
            str: Prefixed passkey with storage key prefix.
        """

        return "{}{}".format(self.get_pass_prefix(), key)

    def get_pass_prefix(self):
        """Passkey prefix getter.

        Returns:
            str: Prefix for storage key.
        """

        return self.get_prefix(self.PASS)

    def get_host_key(self, key):
        """Hostname generator and getter.

        Args:
            key (str): Key to format and return with prefix.

        Returns:
            str: Prefixed hostname with host key prefix.
        """

        return "{}{}".format(self.get_host_prefix(), key)

    def get_host_prefix(self):
        """Hostname prefix getter.

        Returns:
            str: Prefix for hostname key.
        """

        return self.get_prefix(self.HOST)

    def get_jump_key(self, key):
        """Jump generator and getter.

        Args:
            key (str): Key to format and return with prefix.

        Returns:
            str: Prefixed jump authority with jump key prefix.
        """

        return "{}{}".format(self.get_jump_prefix(), key)

    def get_jump_prefix(self):
        """Jump prefix getter.

        Returns:
            str: Prefix for jump key.
        """

        return self.get_prefix(self.JUMP)

    def get_auth_key(self, key):
        """Authority generator and getter.

        Args:
            key (str): Key to format and return with prefix.

        Returns:
            str: Prefixed authority with auth key prefix.
        """

        return "{}{}".format(self.get_auth_prefix(), key)

    def get_auth_prefix(self):
        """Authority prefix getter.

        Returns:
            str: Prefix for authority key.
        """

        return self.get_prefix(self.AUTH)

    def which(self, key):
        """Determine the type of a key.

        Args:
            key (str): Key to determine type.

        Returns:
            str: Type of key if supported, otherwise "unsupported".
        """

        if key.startswith(self.PASS):
            return "storage"
        elif key.startswith(self.AUTH):
            return "authority"
        elif key.startswith(self.HOST):
            return "hostname"
        elif key.startswith(self.JUMP):
            return "jump server"
        return "unsupported"



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import logging


class Log(object):
    """Log wrapper. That's it...
    """

    config = {
        r"format": r"%(message)s"
    }

    @classmethod
    def configure(cls, verbose=False):
        if "DEBUG" in os.environ or verbose:
            cls.config["level"] = logging.DEBUG
        else:
            cls.config["level"] = logging.WARNING
        logging.basicConfig(**cls.config)

    @classmethod
    def text_fmt(cls, message, params=None):
        if isinstance(params, dict):
            for k in params.iterkeys():
                if "{" + k + "}" not in message:
                    raise SystemExit("Unused key {} in text format".format(k))
            return message.format(**params)
        return message

    @classmethod
    def info(cls, message, **kwargs):
        logging.info(cls.text_fmt(message, kwargs))

    @classmethod
    def debug(cls, message, **kwargs):
        logging.debug(cls.text_fmt(message, kwargs))

    @classmethod
    def warn(cls, message, **kwargs):
        logging.warning(cls.text_fmt(message, kwargs))

    @classmethod
    def error(cls, message, **kwargs):
        logging.error(cls.text_fmt(message, kwargs))

    @classmethod
    def fatal(cls, message, **kwargs):
        error = cls.text_fmt(message, kwargs)
        if os.environ.get("DEBUG") == "true":
            throw = Exception
        else:
            throw = SystemExit
        raise throw(error)



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from socket import gethostbyname
from ipaddress import ip_address, IPv4Address, IPv6Address, AddressValueError
from zlib import crc32

from unlocker.util.log import Log


class Authority(object):
    """Object authority holder.

    Defined by https://tools.ietf.org/html/rfc3986#section-3.2

    Arguments:
        MIN_PORT     (int): min supported port number.
        MAX_PORT     (int): max supported port number.
        MAX_USER_LEN (int): max string length for username.
        DELIMITER    (str): authority delimiter used in string representation.
        COMPONENTS   (int): total number of components in a representation.

    Args:
        host    (int): IP address stored as an integer.
        port    (int): Port number of hostname.
        user    (str): Username assigned to hostname.
        scheme  (str): Connection service scheme (has default; optional).
        ip_addr (str): Resolved hostname to IP address (has default value).

    """

    MIN_PORT, MAX_PORT = 1, (2**16)-1
    MAX_USER_LEN = 32
    DELIMITER, COMPONENTS = ":", 3
    HUMAN_READABLE_FORMAT = "{scheme}://{user}@{ipv4}:{port}"
    COMPONENTS_FORMAT = "{host}:{port}:{user}:{scheme}"

    def __init__(self):
        self.host, self.port, self.user = None, None, None
        self.scheme = "tcp"
        self.ip_addr = u"0.0.0.0"

    def get_host(self):
        """Authority host getter.

        Returns:
            int: Hostname as integer.
        """

        return self.host

    def get_host_ip4(self):
        """Authority IPv4 host getter.

        Returns:
            unicode: Hostname as IPv4 address.
        """

        return unicode(IPv4Address(self.host))

    def get_host_ip6(self):
        """Authority IPv6 host getter.

        Returns:
            unicode: Hostname as IPv6 address.
        """

        return unicode(IPv6Address(self.host))

    def set_host(self, host):
        """Authority host setter.

        Args:
            host (str): Hostname or IP address to store.

        Raises:
            Exception: if an invalid hostname is provided.
        """

        if not isinstance(host, (str, unicode)):
            Log.fatal("Invalid host: expected string, got {x}", x=type(host))
        try:
            self.ip_addr = unicode(gethostbyname(host))
        except Exception as e:
            Log.warn("Cannot resolve hostname {h}: {e}", e=str(e), h=host)
        try:
            self.host = int(ip_address(self.ip_addr))
        except Exception as e:
            Log.fatal("Invalid host: {e}", e=str(e))

    def get_port(self):
        """Authority port getter.

        Returns:
            int: Port number.
        """

        return self.port

    def set_port(self, port):
        """Authority port setter.

        Args:
            port (int): Port number of connection.

        Raises:
            Exception: if an invalid port is provided.
        """

        if isinstance(port, (str, unicode)):
            port = int(port)
        if not isinstance(port, int):
            Log.fatal("Invalid port: expected integer, got {x}", x=type(port))
        if port < self.MIN_PORT or port > self.MAX_PORT:
            Log.fatal("Invalid port: out of range {port}", port=port)
        self.port = port

    def get_user(self):
        """Authority user getter.

        Returns:
            str: Connection username.
        """

        return self.user

    def set_user(self, user):
        """Authority user setter.

        Args:
            user (str): Username assigned to hostname.

        Raises:
            Exception: if an invalid user is provided.
        """

        if not isinstance(user, (str, unicode)):
            Log.fatal("Invalid user: expected string, got {x}", x=type(user))
        if len(user) == 0:
            Log.fatal("Invalid user: zero-length string not allowed")
        if len(user) > self.MAX_USER_LEN:
            Log.fatal("Invalid user: max length exceeded {v}", v=len(user))
        self.user = user

    def get_scheme(self):
        """Authority scheme getter.

        Returns:
            str: Connection scheme.
        """

        return self.scheme

    def set_scheme(self, scheme):
        """Authority scheme setter.

        Args:
            scheme (str): Connection service scheme.

        Raises:
            Exception: if an invalid scheme is provided.
        """

        if not isinstance(scheme, (str, unicode)):
            Log.fatal("Invalid scheme: expected string, got {x}",
                      x=type(scheme))
        if len(scheme) == 0:
            Log.fatal("Invalid scheme: zero-length string not allowed")
        self.scheme = scheme

    def read(self, human_readable=False):
        """Read authority representation.

        Returns:
            str: Representation of authority.
        """

        if human_readable:
            return self.HUMAN_READABLE_FORMAT.format(
                user=self.user, ipv4=self.get_host_ip4(), port=self.port,
                scheme=self.scheme)
        return self.COMPONENTS_FORMAT.format(
                host=self.host, port=self.port, user=self.user,
                scheme=self.scheme)

    def signature(self):
        """Find the CRC32 hash of current authority.

        Returns:
            str: Calculated CRC for current authority.
        """

        return self.__class__.sign(self.read())

    def __repr__(self):
        return "[{} {}]".format(self.signature(), self.read(True))

    @classmethod
    def sign(cls, data):
        """Calculate CRC32 hash of given data.

        Args:
            data (str): String data to calculate CRC.

        Returns:
            str: Hex value without 0x of the calculated CRC32.
        """

        if not isinstance(data, (str, unicode)):
            Log.fatal("Cannot calculate CRC for non-string data")
        return hex(crc32(data) & 0xFFFFFFFF)[2:]  # skip 0x

    @classmethod
    def new(cls, host, port, user, scheme=None):
        """Create new authority instance.

        Args:
            host   (str): Hostname or IP address as an integer.
            port   (int): Port number of hostname.
            user   (str): Username assigned to hostname.
            scheme (str): Connection service scheme (optional).

        Raises:
            Exception: If required fields are invalid.

        Returns:
            Authority: An authority instance.
        """

        auth = cls()
        try:
            if scheme is not None:
                auth.set_scheme(scheme)
            auth.set_host(host)
            auth.set_port(port)
            auth.set_user(user)
        except Exception as e:
            Log.fatal("Cannot create new authority: {e}", e=str(e))
        return auth

    @classmethod
    def recover(cls, authority):
        """Reconstruct an authority instance from a string representation.

        Args:
            authority (str): A string representation of a stored authority.

        Raises:
            Exception: If required fields are invalid.

        Returns:
            Authority: An authority instance.
        """

        if authority.count(cls.DELIMITER) != cls.COMPONENTS:
            Log.fatal("Cannot recover from an invalid authority")
        host, port, user, srv = authority.split(cls.DELIMITER, cls.COMPONENTS)
        try:  # fix for non-linux platforms
            host = str(IPv4Address(int(host, 10)))
        except ValueError as e:
            Log.warn("Cannot convert to IP4: {e}", e=str(e))
        except AddressValueError as e:
            Log.fatal("Invalid host address: {e}", e=str(e))
        return cls.new(host, port, user, srv)



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os import path
from getpass import getpass
from base64 import b64encode

from unlocker.util.log import Log


class Passkey(object):
    """Higher-level secrets manager.

    Handlers passkeys from secrets formats and converts between them.

    Arguments:
        passkey (str): The passkey to save or convert.
        passfix (str): The prefix of the supported convertion passkey.
    """

    SUPPORTED_TYPES = {"password": ".", "privatekey": ">"}
    MIN_LEN = 2

    passkey = None

    def __init__(self, auth_type):
        if auth_type not in self.SUPPORTED_TYPES:
            Log.fatal("Unsupported auth type for passkey: {t}", t=auth_type)
        self.__read_method = "read_{}".format(auth_type)
        self.passfix = self.SUPPORTED_TYPES.get(auth_type)

    @classmethod
    def resolve(cls, auth):
        """Returns the final version of a passkey with its appropriate type.

        Args:
            auth (str): Authentification method.

        Raises:
            Exception: If unsupported authentification or cannot resolve pin.

        Returns:
            str: Prefixed passkey to store it later.
        """

        pk = cls(auth)
        pk.read()
        return pk.pin()

    @classmethod
    def copy(cls, secret, unsafe=False):
        """Creates an in-memory copy a passkey.

        Args:
            secret  (str): The secret to unpack and create a copy of it.
            unsafe (bool): Whether to return the copy as unsafe or not.

        Raises:
            Exception: If unsupported secret passkey type found.

        Returns:
            tuple: Passkey type and copy of unsolved (raw) secret.
        """

        if len(secret) <= cls.MIN_LEN:
            Log.fatal("Secret size too small")
        for ptype, prefix in cls.SUPPORTED_TYPES.iteritems():
            if secret[0] == prefix:
                passkey = b64encode(secret[1:]) if not unsafe else secret[1:]
                return ptype, passkey
        Log.fatal("Cannot copy secret: unsupported resolve method")

    def pin(self):
        """Return the prefixed secret with its appropriate type.

        Raises:
            Exception: If unsupported passkey is provided.

        Returns:
            str: Prefixed passkey.
        """

        if self.passkey is None:
            Log.fatal("Passkey has not been read yet")
        return self.passfix + self.passkey

    def save(self, passkey):
        """Save raw passkey "AS IS".

        Args:
            passkey (str): Passkey to save.
        """

        self.passkey = passkey

    def read(self):
        """Autodetect read method and retrieve passkey.

        Raises:
            Exception: If cannot autodetect read method.
        """

        if not hasattr(self, self.__read_method):
            Log.fatal("Unsupported read method: {m}", m=self.read_method)
        self.passkey = getattr(self, self.__read_method)()

    def read_password(self, prompt="Password: "):
        """Get password from user input.

        Args:
            password_prompt (str): Prompt message to display.
        """

        return getpass(prompt)

    def read_privatekey(self, prompt="Path to private key: "):
        """Get private key from user input.

        Args:
            pk_prompt (str): Prompt message to display.

        Raises:
            Exception: If path to private key does not exist or cannot open.
        """

        filepath = raw_input(prompt)
        if not path.exists(filepath):
            Log.fatal("Path to private key does not exists")
        try:
            with open(filepath, "rb") as fd:
                return fd.read()
        except Exception as e:
            Log.fatal("Cannot read private key: {e}", e=str(e))



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from unlocker.manager import Manager
from unlocker.stream import StreamData

from unlocker.util.secret import confidential
from unlocker.util.shell import ShellParser
from unlocker.util.log import Log


@confidential
def unlocker(secrets, args=()):
    """Manage unlocker's keychain.

    Can add, edit, delete and lookup keys.
    All keys stored are compressed and encoded.
    """

    # register secrets on keychain
    Manager.initialize(secrets)
    Log.debug("Preparing to boot...")

    # initialize manager and parse arguments
    mng = Manager(*args)
    mng.call()
    Log.debug("Preparing to exit...")


def read_input():
    """Read input from stdin or shell.

    Some shell arguments can trigger exit before reaching end of function.
    """

    # check if data is piped to unlocked and dump passkey
    if StreamData.read():
        Log.debug("Reading stdin: {data}", data=StreamData.buf_in)
        return StreamData.OPTION, StreamData.parse()

    # initialize shell
    shell = ShellParser()
    opts, args = shell.get_args()
    Log.debug("Running '{o}' with arguments: {a}", o=opts, a=args)

    # return shell arguments
    return opts, vars(args)


def main():
    """Main callable function.

    Configure logging capabilities, read input from shell or stdin (can exit
    fastly) and run unlocker manager.
    """

    # configure log
    Log.configure()
    Log.debug("Running in debug mode...")

    # read input
    args = read_input()

    # run unlocker with input args
    unlocker(args=args)



#!/usr/bin/env python

from setuptools import setup

from unlocker import __version__


_desc = "Unlocker is a keychain and a CLI credentials manager. Useful when " \
        "you use a terminal often than GUI applications for remote " \
        "connections (e.g. databases, SSH, rsync). It can store passwords " \
        "and private keys. It comes with addons to encrypt keychain and " \
        "fast passwordless connect to servers."


setup(name="unlocker",
    packages=[
        "unlocker",
        "unlocker.util",
        "unlocker.data",
    ],
    include_package_data=True,
    entry_points = {
        "console_scripts": [
            "unlocker = unlocker.bootstrap:main"
        ]
    },
    install_requires=[
        "ipaddress==1.0.17",
        "click==6.7"
    ],
    test_suite="tests",
    version=__version__,
    description="CLI credentials manager",
    long_description=_desc,
    author="Alexandru Catrina",
    author_email="alex@codeissues.net",
    license="MIT",
    url="https://github.com/lexndru/unlocker",
    download_url="https://github.com/lexndru/unlocker/archive/v{}.tar.gz".format(__version__),
    keywords=["credentials manager", "keychain", "remote connection"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Education",
        "Topic :: System :: Networking",
        "Topic :: System :: Shells",
        "Topic :: System :: Systems Administration",
        "Topic :: Terminals",
        "Topic :: Utilities",
        "Topic :: Internet",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Unix Shell",
        "Operating System :: POSIX",
    ],
)



#!/usr/bin/env python
#
# Copyright 2018 Alexandru Catrina
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from unlocker.util.log import Log


class Service(object):
    """Services helper class.

    Useful to map missing ports on some services with default known ports.
    """

    services = (
        # scheme    port
        ("http",    80),
        ("https",   443),
        ("kafka",   9092),
        ("mongo",   27017),
        ("mysql",   3306),
        ("neo4j",   7474),
        ("pgql",    5432),
        ("redis",   6379),
        ("rsync",   873),
        ("smtp",    25),
        ("ssh",     22),
        ("ftp",     21),
    )

    @classmethod
    def find_port(cls, service):
        """Find port number for service.

        Args:
            service (str): Service scheme name.

        Raises:
            Exception: If service is not supported.

        Return:
            int: Port number for service.
        """

        for srv, port in cls.services:
            if srv == service:
                return port
        Log.fatal("Unsupported service {s}", s=service)

    @classmethod
    def find_scheme(cls, port):
        """Find scheme for service by port number.

        Args:
            port (int): Service port number.

        Return:
            str: Scheme for service or None.
        """

        for scheme, srv_port in cls.services:
            if int(port) == int(srv_port):
                return scheme
        return None
